/*****************************************************************************
 *
 * jack.c
 *
 * PHASEX:  [P]hase [H]armonic [A]dvanced [S]ynthesis [EX]periment
 *
 * Copyright (C) 1999-2007 William Weston <weston@sysex.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************/
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <jack/jack.h>
#include "phasex.h"
#include "jack.h"
#include "engine.h"
#include "settings.h"


jack_client_t *client;

jack_port_t *input_port1;
jack_port_t *input_port2;

jack_port_t *output_port1;
jack_port_t *output_port2;

pthread_mutex_t buffer_mutex		= PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t sample_rate_mutex	= PTHREAD_MUTEX_INITIALIZER;

pthread_cond_t buffer_has_data		= PTHREAD_COND_INITIALIZER;
pthread_cond_t buffer_has_space		= PTHREAD_COND_INITIALIZER;
pthread_cond_t sample_rate_cond		= PTHREAD_COND_INITIALIZER;

jack_default_audio_sample_t input_buffer1[PHASEX_MAX_BUFSIZE];
jack_default_audio_sample_t input_buffer2[PHASEX_MAX_BUFSIZE];

jack_default_audio_sample_t output_buffer1[PHASEX_MAX_BUFSIZE];
jack_default_audio_sample_t output_buffer2[PHASEX_MAX_BUFSIZE];

int buffer_full		= 0;
int buffer_needed	= PHASEX_MAX_BUFSIZE;
int buffer_size		= PHASEX_MAX_BUFSIZE;
int buffer_read_index	= 0;
int buffer_write_index	= 0;

int sample_rate		= 0;
int sample_rate_mode	= SAMPLE_RATE_NORMAL;
float_t f_sample_rate	= 0.0;
float_t nyquist_freq	= 22050.0;
float_t wave_period	= F_WAVEFORM_SIZE / 44100.0;

char *jack_inputs	= NULL;
char *jack_outputs	= NULL;


/*****************************************************************************
 *
 * PROCESS_BUFFER()
 *
 * callback for jack to read audio data from ringbuffer
 * jack uses a realtime priority thread for this
 *
 *****************************************************************************/
int 
process_buffer(jack_nframes_t nframes, void *arg) {
    jack_default_audio_sample_t *in1;
    jack_default_audio_sample_t *in2;
    jack_default_audio_sample_t *out1;
    jack_default_audio_sample_t *out2;


    /* Get the input and output buffers for audio data */
    in1  = jack_port_get_buffer (input_port1,  nframes);
    in2  = jack_port_get_buffer (input_port2,  nframes);
    out1 = jack_port_get_buffer (output_port1, nframes);
    out2 = jack_port_get_buffer (output_port2, nframes);

    /* check for global shutdown so we don't get hung on mutex */
    if (shutdown) {
	memset (out1, 0, (sizeof (jack_default_audio_sample_t) * nframes));
	memset (out2, 0, (sizeof (jack_default_audio_sample_t) * nframes));
	return 0;
    }

    /* this process callback needs the number of frames that jack calls for */
    buffer_needed = nframes;

    /* copy buffers and update index */
    memcpy (&(input_buffer1[buffer_read_index]), in1,   sizeof (jack_default_audio_sample_t) * nframes);
    memcpy (&(input_buffer2[buffer_read_index]), in2,   sizeof (jack_default_audio_sample_t) * nframes);
    memcpy (out1, &(output_buffer1[buffer_read_index]), sizeof (jack_default_audio_sample_t) * nframes);
    memcpy (out2, &(output_buffer2[buffer_read_index]), sizeof (jack_default_audio_sample_t) * nframes);
    buffer_read_index += nframes;
    buffer_read_index %= buffer_size;

    /* Update buffer full and buffer needed counters */
    buffer_full   -= nframes;
    buffer_needed = buffer_size;

    /* Signal the engine that there's space again */
    if (pthread_mutex_trylock (&buffer_mutex) == 0) {
	pthread_cond_broadcast (&buffer_has_space);
	pthread_mutex_unlock (&buffer_mutex);
    }

    return 0;
}


/*****************************************************************************
 *
 * JACK_BUFSIZE_HANDLER()
 *
 * Gets called when jack sets or changes buffer size.
 *
 *****************************************************************************/
int 
jack_bufsize_handler(jack_nframes_t nframes, void *arg) {
    /* Begin buffer critical section */
    pthread_mutex_lock (&buffer_mutex);

    /* Make sure buffer doesn't get overrun */
    if (nframes > PHASEX_MAX_BUFSIZE) {
	fprintf (stderr, "JACK requested buffer size:  %d.  Max is:  %d.\n",
		 nframes, PHASEX_MAX_BUFSIZE);
	phasex_shutdown ("Buffer size exceeded.  Exiting...\n");
    }
    buffer_size   = nframes;
    buffer_full   = buffer_size;
    buffer_needed = 0;

    pthread_mutex_unlock (&buffer_mutex);
    /* End buffer critical section */

    if (debug) {
	fprintf (stderr, "JACK requested buffer size:  %d\n", nframes);
    }

    return 0;
}


/*****************************************************************************
 *
 * JACK_SAMPLERATE_HANDLER()
 *
 * Gets called when jack sets or changes buffer size.
 *
 *****************************************************************************/
int 
jack_samplerate_handler(jack_nframes_t nframes, void *arg) {

    if (debug) {
	fprintf (stderr, "JACK requested sample rate:  %d\n", nframes);
    }

    /* if jack requests a zero value, just use what we already have */
    if (nframes == 0) {
	return 0;
    }

    /* scale sample rate depending on mode */
    switch (sample_rate_mode) {
    case SAMPLE_RATE_UNDERSAMPLE:
	nframes /= 2;
	break;
    case SAMPLE_RATE_NORMAL:
	break;
    case SAMPLE_RATE_OVERSAMPLE:
	nframes *= 2;
	break;
    }

    pthread_mutex_lock (&sample_rate_mutex);
    if (nframes == sample_rate) {
	pthread_mutex_unlock (&sample_rate_mutex);
	return 0;
    }

    /* Changing sample rate currently not supported */
    if ((sample_rate > 0) && (sample_rate != nframes)) {
	/* TODO: rebuild _all_ samplerate dependent tables and vars here */
	phasex_shutdown ("Unable to change sample rate.  Exiting...\n");
    }

    /* First time setting sample rate */
    if (sample_rate == 0) {
	sample_rate    = nframes;
	f_sample_rate  = (float_t)sample_rate;
	nyquist_freq   = (float_t)(sample_rate / 2);
	wave_period    = (float_t)(F_WAVEFORM_SIZE / (double)sample_rate);

	if (debug) {
	    fprintf (stderr, "Internal sample rate:  %d\n", nframes);
	}

	/* now that we have the sample rate, signal anyone else who needs to know */
	pthread_cond_broadcast (&sample_rate_cond);
    }

    pthread_mutex_unlock (&sample_rate_mutex);
    return 0;
}


/*****************************************************************************
 *
 * JACK_SHUTDOWN_HANDLER()
 *
 * Gets called when JACK shuts down or closes client.
 *
 *****************************************************************************/
void 
jack_shutdown_handler(void *arg) {
    if (debug) {
	fprintf (stderr, "jack_shutdown_handler()\n");
    }
    phasex_shutdown ("JACK shut down client.\n");
}


/*****************************************************************************
 *
 * JACK_XRUN_HANDLER()
 *
 * Gets called when jack detects an xrun event.
 *
 *****************************************************************************/
int 
jack_xrun_handler(void *arg) {
    if (debug) {
	fprintf (stderr, "JACK xrun detected.\n");
    }
    return 0;
}


/*****************************************************************************
 * 
 * JACK_CLEANUP()
 *
 * Handles JACK disconnection at thread cleanup time.
 *
 *****************************************************************************/
void 
jack_cleanup(void *arg) {
    if (debug) {
	fprintf (stderr, "Entering jack_cleanup()...\n");
    }
    /* deactivate and close client */
    if (client != NULL) {
#ifdef JACK_DEACTIVATE_BEFORE_CLOSE
	jack_deactivate (client);
#endif
	jack_client_close (client);
	client = NULL;
    }
    /* make sure the engine thread doesn't get hung */
    if (pthread_mutex_trylock (&buffer_mutex) == 0) {
	buffer_full   = 0;
	buffer_needed = buffer_size;
	pthread_cond_broadcast (&buffer_has_space);
	pthread_mutex_unlock (&buffer_mutex);
    }
}


/*****************************************************************************
 *
 * JACK_THREAD()
 *
 * Setup for becoming a jack client and attaching to playback ports
 *
 *****************************************************************************/
void * 
jack_thread(void *arg) {
    const char		**ports;
    const char		*client_name = "phasex";
    const char		*server_name = NULL;
    char		*p;
    char		*q;
    jack_options_t	options = JackNullOption;
    jack_status_t	status;
    int			j;


    /* open a client connection to the JACK server */
    client = jack_client_open (client_name, options, &status, server_name);
    if (client == NULL) {
	if (status & JackServerFailed) {
	    fprintf (stderr, "Unable to connect to JACK server.\n");
	}
	fprintf (stderr, "JACK Status = 0x%2.0x\n", status);
	phasex_shutdown ("Unable to open JACK client.\n");
    }

    /* set thread cleanup handler for more graceful shutdown disconnects */
    pthread_cleanup_push (jack_cleanup, NULL);

    /* check for duplicate client name */
    if (status & JackNameNotUnique) {
	if (debug) {
	    fprintf (stderr, "Unique JACK client name '%s' already assigned.\n", client_name);
	}
	client_name = jack_get_client_name (client);
	if (debug) {
	    fprintf (stderr, "JACK assigned unique client name '%s'.\n", client_name);
	}
    }
    if (status & JackServerStarted) {
	fprintf (stderr, "JACK server started\n");
    }

    /* report realtime scheduling in JACK */
    if (debug) {
	if (jack_is_realtime (client)) {
	    fprintf (stderr, "JACK is running with realtime scheduling.\n");
	}
	else {
	    fprintf (stderr, "JACK is running without realtime scheduling.\n");
	}
    }

    /* get sample rate early and notify other threads */
    sample_rate = jack_get_sample_rate (client);

    if (debug) {
	fprintf (stderr, "JACK sample rate:  %d\n", sample_rate);
    }

    /* scale sample rate depending on mode */
    switch (sample_rate_mode) {
    case SAMPLE_RATE_UNDERSAMPLE:
	sample_rate /= 2;
	break;
    case SAMPLE_RATE_OVERSAMPLE:
	sample_rate *= 2;
	break;
    }

    /* set samplerate related vars */
    f_sample_rate = (float_t)sample_rate;
    nyquist_freq  = (float_t)(sample_rate / 2);
    wave_period   = (float_t)(F_WAVEFORM_SIZE / (double)sample_rate);

    if (debug) {
	fprintf (stderr, "Internal sample rate:  %d\n", sample_rate);
    }

    /* callback for setting our sample rate when jack tells us to */
    jack_set_sample_rate_callback (client, jack_samplerate_handler, 0);

    /* now that we have the sample rate, signal anyone else who needs to know */
    pthread_mutex_lock (&sample_rate_mutex);
    pthread_cond_broadcast (&sample_rate_cond);
    pthread_mutex_unlock (&sample_rate_mutex);

    /* get buffer size */
    if ((buffer_size = jack_get_buffer_size (client)) > PHASEX_MAX_BUFSIZE) {
    	fprintf (stderr, "JACK requested buffer size:  %d.  Max is:  %d.\n",
		 buffer_size, PHASEX_MAX_BUFSIZE);
    	phasex_shutdown ("JACK buffer size exceeded.  Exiting...\n");
    }
    buffer_full = buffer_size;
    buffer_needed = 0;
    if (debug) {
	fprintf (stderr, "JACK output buffer size:  %d.\n", buffer_size);
    }

    /* callback for setting our buffer size when jack tells us to */
    jack_set_buffer_size_callback (client, jack_bufsize_handler, 0);

    /* callback for when jack shuts down */
    jack_on_shutdown (client, jack_shutdown_handler, 0);

    /* callback for dealing with xruns */
    jack_set_xrun_callback (client, jack_xrun_handler, 0);

    /* create ports */
    input_port1 = jack_port_register (client, "in_1",
				      JACK_DEFAULT_AUDIO_TYPE,
				      JackPortIsInput, 0);
    input_port2 = jack_port_register (client, "in_2",
				      JACK_DEFAULT_AUDIO_TYPE,
				      JackPortIsInput, 0);

    output_port1 = jack_port_register (client, "out_1",
				       JACK_DEFAULT_AUDIO_TYPE,
				       JackPortIsOutput, 0);
    output_port2 = jack_port_register (client, "out_2",
				       JACK_DEFAULT_AUDIO_TYPE,
				       JackPortIsOutput, 0);

    if ( (output_port1 == NULL) || (output_port2 == NULL) ||
	 (input_port1 == NULL) || (input_port2 == NULL) ) {
	phasex_shutdown ("JACK has no output ports available.\n");
    }

    /* callback for jack to grab audio data */
    jack_set_process_callback (client, process_buffer, 0);

    /* wait for the engine to be set up */
    pthread_mutex_lock (&engine_ready_mutex);
    if (!engine_ready) {
	pthread_cond_wait (&engine_ready_cond, &engine_ready_mutex);
    }
    pthread_mutex_unlock (&engine_ready_mutex);

    /* activate client (callbacks start, so everything else needs to be ready) */
    if (jack_activate (client)) {
	phasex_shutdown ("Unable to activate JACK client.\n");
    }

    /* connect ports.  in/out is from server perspective */

    /* By default, connect PHASEX output to first two JACK hardware playback ports found */
    if (setting_jack_autoconnect) {
	ports = jack_get_ports (client, NULL, NULL,
				JackPortIsPhysical | JackPortIsInput);
	if (ports == NULL) {
	    fprintf (stderr, "No physical JACK playback ports available.\n");
	}
	else {
	    if (jack_connect (client, jack_port_name (output_port1), ports[0])) {
		fprintf (stderr, "Unable to connect PHASEX output ports to JACK input.\n");
	    }
	    if (jack_connect (client, jack_port_name (output_port2), ports[1])) {
		fprintf (stderr, "Unable to connect PHASEX output ports to JACK input.\n");
	    }
	}
    }

    /* otherwise, connect to ports that match -o command line flag */
    else {
	ports = jack_get_ports (client, NULL, NULL, JackPortIsInput);
	if (ports == NULL) {
	    fprintf (stderr, "No JACK playback ports available.\n");
	}
	else {
	    p = jack_outputs;
	    if (p != NULL) {
		if ((q = index (p, ',')) != NULL) {
		    *q = '\0';
		    q++;
		}
		else {
		    q = p;
		}
		for (j = 0; ports[j] != NULL; j++) {
		    if (strstr (ports[j], p) != NULL) {
			if (jack_connect (client, jack_port_name(output_port1), ports[j])) {
			    fprintf (stderr, "Unable to connect PHASEX output ports to JACK input.\n");
			}
			else if (debug) {
			    fprintf (stderr, "JACK connected PHASEX out_1 to %s.\n", ports[j]);
			}
			p = "_no_port_";
		    }
		    if (strstr (ports[j], q) != NULL) {
			if (jack_connect (client, jack_port_name(output_port2), ports[j])) {
			    fprintf (stderr, "Unable to connect PHASEX output ports to JACK input.\n");
			}
			else if (debug) {
			    fprintf (stderr, "JACK connected PHASEX out_2 to %s.\n", ports[j]);
			}
			q = "_no_port_";
		    }
		}
	    }
	}
    }
    free (ports);

    /* By default, connect PHASEX input to first two JACK hardware capture ports found */
    if (setting_jack_autoconnect) {
	ports = jack_get_ports (client, NULL, NULL,
				JackPortIsPhysical | JackPortIsOutput);
	if (ports == NULL) {
	    fprintf (stderr, "No physical JACK capture ports available.\n");
	}
	else {
	    if (jack_connect (client, ports[0], jack_port_name (input_port1))) {
		fprintf (stderr, "Unable to connect PHASEX input ports to JACK output.\n");
	    }
	    if (jack_connect (client, ports[1], jack_port_name (input_port2))) {
		fprintf (stderr, "Unable to connect PHASEX input ports to JACK output.\n");
	    }
	}
    }

    /* otherwise, connect to ports that match -i command line flag */
    else {
	ports = jack_get_ports (client, NULL, NULL, JackPortIsOutput);
	if (ports == NULL) {
	    fprintf (stderr, "No JACK capture ports available.\n");
	}
	else {
	    p = jack_inputs;
	    if (p != NULL) {
		if ((q = index (p, ',')) != NULL) {
		    *q = '\0';
		    q++;
		}
		else {
		    q = p;
		}
		for (j = 0; ports[j] != NULL; j++) {
		    if (strstr (ports[j], p) != NULL) {
			if (jack_connect (client, ports[j], jack_port_name (input_port1))) {
			    fprintf (stderr, "Unable to connect PHASEX input ports to JACK output.\n");
			}
			else if (debug) {
			    fprintf (stderr, "JACK connected %s to PHASEX in_1.\n", ports[j]);
			}
			p = "_no_port_";
		    }
		    if (strstr (ports[j], q) != NULL) {
			if (jack_connect (client, ports[j], jack_port_name (input_port2))) {
			    fprintf (stderr, "Unable to connect PHASEX input ports to JACK output.\n");
			}
			else if (debug) {
			    fprintf (stderr, "JACK connected %s to PHASEX in_2.\n", ports[j]);
			}
			q = "_no_port_";
		    }
		}
	    }
	}
    }
    free (ports);

    /* report latency and total latency */
    if (debug) {
	fprintf (stderr, "JACK latency:  %d\n",
		 jack_port_get_latency (output_port1));
	fprintf (stderr, "JACK total latency:  %d\n",
		 jack_port_get_total_latency (client, output_port1));
    }

    /* sit here and wait */
    while (1) {
	pthread_testcancel ();
	sleep (1);
    }

    /* exit so we can use the cleanup handler */
    pthread_exit (NULL);

    /* make the compiler happy */
    pthread_cleanup_pop (1);
    return NULL;
}
